@@ -3,8 +3,7 @@
 from __future__ import annotations
 
 import asyncio
-from collections.abc import Callable
-from functools import partial
+from functools import cached_property, partial
 import ipaddress
 import logging
 import os
@@ -16,6 +15,7 @@ import weakref
 from aiohttp import web
 
 from matter_server.server.helpers.custom_web_runner import MultiHostTCPSite
+from matter_server.server.helpers.paa_certificates import fetch_certificates
 
 from ..common.const import SCHEMA_VERSION
 from ..common.errors import VersionMismatch
@@ -30,12 +30,18 @@ from ..common.models import (
     ServerInfoMessage,
 )
 from ..server.client_handler import WebsocketClientHandler
-from .const import DEFAULT_PAA_ROOT_CERTS_DIR, MIN_SCHEMA_VERSION
+from .const import (
+    DEFAULT_PAA_ROOT_CERTS_DIR,
+    MIN_SCHEMA_VERSION,
+)
 from .device_controller import MatterDeviceController
 from .stack import MatterStack
 from .storage import StorageController
 from .vendor_info import VendorInfo
 
+if TYPE_CHECKING:
+    from collections.abc import Callable
+
 DASHBOARD_DIR = Path(__file__).parent.joinpath("../dashboard/").resolve()
 DASHBOARD_DIR_EXISTS = DASHBOARD_DIR.exists()
 
@@ -102,6 +108,7 @@ class MatterServer:
         listen_addresses: list[str] | None = None,
         primary_interface: str | None = None,
         paa_root_cert_dir: Path | None = None,
+        enable_test_net_dcl: bool = False,
     ) -> None:
         """Initialize the Matter Server."""
         self.storage_path = storage_path
@@ -114,20 +121,28 @@ class MatterServer:
             self.paa_root_cert_dir = DEFAULT_PAA_ROOT_CERTS_DIR
         else:
             self.paa_root_cert_dir = Path(paa_root_cert_dir).absolute()
+        self.enable_test_net_dcl = enable_test_net_dcl
         self.logger = logging.getLogger(__name__)
         self.app = web.Application()
         self.loop: asyncio.AbstractEventLoop | None = None
         # Instantiate the Matter Stack using the SDK using the given storage path
         self.stack = MatterStack(self)
-        # Initialize our (intermediate) device controller which keeps track
-        # of Matter devices and their subscriptions.
-        self.device_controller = MatterDeviceController(self)
         self.storage = StorageController(self)
         self.vendor_info = VendorInfo(self)
         # we dynamically register command handlers
         self.command_handlers: dict[str, APICommandHandler] = {}
+        self._device_controller: MatterDeviceController | None = None
         self._subscribers: set[EventCallBackType] = set()
-        self._register_api_commands()
+        if MIN_SCHEMA_VERSION > SCHEMA_VERSION:
+            raise RuntimeError(
+                "Minimum supported schema version can't be higher than current schema version."
+            )
+
+    @cached_property
+    def device_controller(self) -> MatterDeviceController:
+        """Return the main Matter device controller."""
+        assert self._device_controller
+        return self._device_controller
 
     async def start(self) -> None:
         """Start running the Matter server."""
@@ -140,9 +155,23 @@ class MatterServer:
         self.loop = asyncio.get_running_loop()
         self.loop.set_exception_handler(_global_loop_exception_handler)
         self.loop.set_debug(os.environ.get("PYTHONDEBUG", "") != "")
-        await self.device_controller.initialize(self.paa_root_cert_dir)
+
+        # (re)fetch all PAA certificates once at startup
+        # NOTE: this must be done before initializing the controller
+        await fetch_certificates(
+            self.paa_root_cert_dir,
+            fetch_test_certificates=self.enable_test_net_dcl,
+            fetch_production_certificates=True,
+        )
+
+        # Initialize our (intermediate) device controller which keeps track
+        # of Matter devices and their subscriptions.
+        self._device_controller = MatterDeviceController(self, self.paa_root_cert_dir)
+        self._register_api_commands()
+
+        await self._device_controller.initialize()
         await self.storage.start()
-        await self.device_controller.start()
+        await self._device_controller.start()
         await self.vendor_info.start()
         mount_websocket(self, "/ws")
         self.app.router.add_route("GET", "/info", self._handle_info)
@@ -204,15 +233,16 @@ class MatterServer:
     @api_command(APICommand.SERVER_INFO)
     def get_info(self) -> ServerInfoMessage:
         """Return (version)info of the Matter Server."""
-        assert self.device_controller.compressed_fabric_id is not None
+        assert self._device_controller
+        assert self._device_controller.compressed_fabric_id is not None
         return ServerInfoMessage(
             fabric_id=self.fabric_id,
-            compressed_fabric_id=self.device_controller.compressed_fabric_id,
+            compressed_fabric_id=self._device_controller.compressed_fabric_id,
             schema_version=SCHEMA_VERSION,
             min_supported_schema_version=MIN_SCHEMA_VERSION,
             sdk_version=chip_clusters_version(),
-            wifi_credentials_set=self.device_controller.wifi_credentials_set,
-            thread_credentials_set=self.device_controller.thread_credentials_set,
+            wifi_credentials_set=self._device_controller.wifi_credentials_set,
+            thread_credentials_set=self._device_controller.thread_credentials_set,
         )
 
     @api_command(APICommand.SERVER_DIAGNOSTICS)
@@ -232,7 +262,7 @@ class MatterServer:
             if asyncio.iscoroutinefunction(callback):
                 asyncio.create_task(callback(evt, data))
             else:
-                self.loop.call_soon_threadsafe(callback, evt, data)
+                self.loop.call_soon(callback, evt, data)
 
     def scope_ipv6_lla(self, ip_addr: str) -> str:
         """Scope IPv6 link-local addresses to primary interface.
@@ -273,9 +303,9 @@ class MatterServer:
 
     def _register_api_commands(self) -> None:
         """Register all methods decorated as api_command."""
-        for cls in (self, self.device_controller, self.vendor_info):
+        for cls in (self, self._device_controller, self.vendor_info):
             for attr_name in dir(cls):
-                if attr_name.startswith("__"):
+                if attr_name.startswith("_"):
                     continue
                 val = getattr(cls, attr_name)
                 if not hasattr(val, "api_cmd"):
